/*

Grid 2D library by ACWraith

////////////////////////////////////////////////////////////////////////////////
LICENSE:	You may use this library for profit or non-profit.
			Credit is appreciated, but not required.

	If you need a more formal license, consider it to be:
		Creative Commons Attribution 3.0 Unported (CC BY 3.0)
		http://creativecommons.org/licenses/by/3.0/

	CC licenses generally are't recommended for source code.
	However, I figure I can keep the rules simple by using the
	same license as any other resources I may make open.  (For
	instance, Fooldom Come subscribers get free reign over the sound
	and	graphics from some of my games.)  Contact me if you need a
	more specialized license:  ACWraith@gmail.com

////////////////////////////////////////////////////////////////////////////////

DESCRIPTION:
	This library offers a datum for handling 2D grids of items.  What those
	items are does not matter.  They can even be null values.

	Procs are included to aid with pathfinding.  Flipping, shifting and rotating
	transformations are also available.

////////////////////////////////////////////////////////////////////////////////

grid2d_Grid

	New(Width=1,Height=1,DefaultValue=null)
		Width and Height are the maximum X and Y dimensions of the grid.
		DefaultValue is what every value in the list is initialized to.
		If Width is a grid2d_Grid, this becomes a copy constructor and
		the other arguments are ignored.

	grid2d_CopyIntoSelf(grid2d_Grid/Grid)
		Turns the grid into a copy of Grid.

	grid2d_Reset(Width=1,Height=1,DefaultValue=null)
		Sets the grid dimensions to Width by Height and makes DefaultValue
		the item stored in every cell.  If Width is a grid2d_Grid, its values
		will be used and the other arguments will be ignored.
		Returns:
			0 on failure
			negative if the result is a copy.
			positive if the result is not a copy


	//==========================================================================
	COORDINATES

	Grids are divided into cells labeled by X and Y coordinates.

		grid2d_GetMaxX()
			Returns the width of the grid.  The minimum is always 1.

		grid2d_GetMaxY()
			Returns the height of the grid.  The minimum is always 1.

		grid2d_HasCoordinates(X,Y)
			Returns whether or not the grid contains the coordinates X,Y.



	Coordinates can be referenced if they are atoms on the map.

		grid2d_GetDir(FromX,FromY,ToX,ToY)
			Returns the direction from the coordinates FromX,FromY to ToX,ToY.
			By default, this is implemented to return NORTH, EAST, SOUTH and WEST
			only when a coordinate is shared (just like get_dir()).

		grid2d_GetDist(X1,Y1,X2,Y2)
			Returns the distance between the coordinates X1,Y1 and X2,Y2.
			Returns negative on failure.

		grid2d_GetXInDir(X,Y,Dir,NumSteps=1)
			Returns the horizontal coordinate of the cell NumSteps in the
			direction of Dir from the cell X,Y.

		grid2d_GetYInDir(X,Y,Dir,NumSteps=1)
			Returns the vertical coordinate of the cell NumSteps in the
			direction of Dir from the cell X,Y.

		There are also some utility procs hidden in dirs.dm for those interested.



	//==========================================================================
	STORAGE

	Items can be stored at each coordinate.  What the items are doesn't matter.
	For instance, null values and lists are perfectly valid.

		grid2d_GetCoordinates(Item)
			Returns a list of the coordinates of every position Item is stored.
			It will look something like: list(X1,Y1,X2,Y2,X3,Y3...)

		grid2d_GetItem(X,Y)
			Returns whatever is stored in the grid at the coordinates X,Y.

		grid2d_GetItems(JustOnce=FALSE)
			Returns a list containing each item stored in the grid.  If JustOnce,
			multiple appearances of the same item will be filtered out of the result.

		grid2d_GetNumOccurances(Item)
			Returns the number of times Item is stored in the grid.

		grid2d_GetNthX(Item,N=1)
			Returns the X coordinate of the Nth occurance of item.  N=1 is the
			first item and N=0 is the last.  Negative values count away from the
			last item.  The occurances are in the order of those returned by
			grid2d_GetCoordinates().

		grid2d_GetNthY(Item,N=1)
			Returns the Y coordinate of the Nth occurance of item.  N=1 is the
			first item and N=0 is the last.  Negative values count away from the
			last item.  The occurances are in the order of those returned by
			grid2d_GetCoordinates().

		grid2d_HasItem(Item)
			Returns whether or not Item is anywhere in the grid.

		grid2d_PlaceItem(Item,X,Y)
			Stores Item at the coordinates X,Y and returns TRUE on success.

		grid2d_Replace(OldItem,Replacement)
			This puts Replacement wherever OldItem is stored and returns the
			number of times it happens

		grid2d_Swap(X1,Y1,X2,Y2)
			Swaps the positions of the objects located at X1,Y1 and X2,Y2.
			Returns TRUE on success.


	//==========================================================================
	COMPARISONS

		Procs are included for the matching of items contained in grids.  This
		is merely a comparison using ==.

		grid2d_GetItemsMatchedByXY(grid2d_Grid/Grid,JustOnce=FALSE)
			Returns a list containing the items listed in the same coordinates
			of the src grid and	Grid.  If JustOnce, multiple appearances of the
			same item will be filtered out of the result.

		grid2d_GetMatchingItemXYs(grid2d_Grid/Grid,JustOnce=FALSE)
			Returns a list of coordinates which have matching items in both the
			src grid and Grid.  If JustOnce, only the first X and Y of each
			item found will be in the result.

		grid2d_XYItemsAllMatch(grid2d_Grid/Grid)
			Compares the src grid with Grid.  Returns true if the items in every
			cell of the smaller grid2d_Grid are in the same positions in both grids.


	//==========================================================================
	FLIPS

	Flipping performs the following transformations depending on the
	direction of the flip.

		origin		N/S		E/W		Ne/Sw	Se/Nw

		ABC			GHI		CBA		ADG		IFC
		DEF			DEF		FED		BEH		HEB
		GHI			ABC		IHG		CFI		GDA

	Diagonally flipping a grid with an unequal width and length will fail.

		grid2d_GetFlipX(X,Y,Dir)
			Find the horizontal coordinate of a cell with coordinate X as if the
			grid was flipped in direction Dir.  Returns 0 on failure.

		grid2d_GetFlipY(X,Y,Dir)
			Find the vertical coordinate of a cell with coordinate Y as if the
			grid was flipped in direction Dir.  Returns 0 on failure.

		grid2d_GetFlipGrid(Dir)
			Returns a grid flipped in direction Dir.  Returns null on failure.



	//==========================================================================
	SHIFTS

	Shifting moves multiple items in the grid along set paths.  Coordinates
	are assumed to wrap across opposite edges.

		grid2d_Shift(OriginX=1,OriginY=1,AmountX=0,AmountY=0)
			Moves the item at OriginX,OriginY the specified horizontal and vertical
			amounts over.  Items in cells along the line formed will be similarly moved.
			However, it will skip the spaces between offsets.

			Examples:

				grid2d_Shift(2,2,1,0)		grid2d_Shift(2,2,0,1)

					ABCD		ABCD		ABCD		AFCD
					EFGH	=>	EFGH		EFGH	=>	EJGH
					IJKL		LIJK		IJKL		INKL
					MNOP		MNOP		MNOP		MBOP

				grid2d_Shift(2,2,2,0)		grid2d_Shift(2,2,0,2)

					ABCD		ABCD		ABCD		AJCD
					EFGH	=>	EFGH		EFGH	=>	EFGH
					IJKL		ILKJ		IJKL		IBKL
					MNOP		MNOP		MNOP		MNOP

				grid2d_Shift(2,2,1,1)		grid2d_Shift(2,2,1,2)

					ABCD		ABCG		ABCD		LBJD
					EFGH	=>	EFJH		EFGH	=>	EFGH
					IJKL		IMKL		IJKL		IAKC
					MNOP		DNOP		MNOP		MNOP

		grid2d_ShiftDir(OriginX=1,OriginY=1,Dir=0,Amount=1)
			Moves the item at OriginX,OriginY the specified amount in direction Dir.
			Items in cells along the line formed will be similarly moved.
			Unlike, grid2d_Shift(), this does not skip the spaces between.
			The entire row, column or diagonal (remember the wrapping) will move.

		grid2d_ShiftPath(list/PathCoords,Amount=1)
			Moves the items at the given coordinates to the Amount next coordinates
			along the list.  Items at opposite ends of the list wrap.  In other
			words this moves items along an arbitrary path.


	//==========================================================================
	TURNS

	Turning performs the following transformations based on the clockwise
	angle.  That angle is expected to be a multiple of 90 degrees.

		initial		90 or -270		180 or -180		270 or -90

		ABC				GDA				IHG				CFI
		DEF				HEB				FED				BEH
		GHI				IFC				CBA				ADG

	If the dimensions of the grid are not equal, they might be swapped
	in the result.

		initial		90 or -270		180 or -180		270 or -90

						DA								CF
		ABC				EB				FED				BE
		DEF				FC				CBA				AD

		grid2d_GetTurnX(X,Y,Angle)
			Get the horizontal coordinate of the X,Y cell as if the grid was turned
			Angle degrees.  Returns 0 on failure.

		grid2d_GetTurnY(X,Y,Angle)
			Get the vertical coordinate of the X,Y cell as if the grid was turned
			Angle degrees.  Returns 0 on failure.

		grid2d_GetTurnGrid(Angle)
			Returns a grid turned clockwise by Angle degrees.  Returns null on
			failure.



	//==========================================================================
	WRAPPING

	Wrapping brings coordinates in bounds as if they exited one side just
	to enter the opposite side.  By default, wrapping is off.  However, it can
	be handled internally automatically and procs are included to do it manually.

		grid2d_DoesWrapX()
			Returns whether or not cells on the EAST wrap to the WEST and vice versa.

		grid2d_DoesWrapY()
			Returns whether or not cells on the SOUTH wrap to the NORTH and vice versa.

		grid2d_GetWrapX(X)
			Find the value of X as if it exited through the WEST or EAST sides
			and came back in through the EAST and WEST sides respectively.
			Returns 0 on failure.

		grid2d_GetWrapY(Y)
			Find the value of Y as if it exited through the SOUTH or NORTH sides
			and came back in through the NORTH and SOUTH sides respectively.
			Returns 0 on failure.

		grid2d_SetWrap(X=FALSE,Y=FALSE)
			Sets whether or not cells on the edges wrap across the horizontal
			and/or vertical axes.



	//==========================================================================
	NEIGHBORS

	Derivative types might want to restrict what cells are considered neighbors.

		grid2d_AreNeighbors(X1,Y1,X2,Y2)
			Returns TRUE is the cells at X1,Y1 and X2,Y2 are right next to each other.
			If the directions are set to not reciprocate with SetNeighborDirs() then
			this only returns TRUE if X1,Y1 acknowledges X2,Y2 as a neighbor and
			not the other way around.

		grid2d_GetNeighbors(X,Y)
			Returns a list of coordinates representing the cells surrounding X,Y.

		grid2d_GetNeighborDirs()
			Returns the list of directions cells are acknowledged to have
			neighbors in.  By default, this is the 8 standard directions.

		grid2d_IsNeighborDir(Dir)
			Returns whether or not cells acknowledge neighbors in direction Dir.

		grid2d_SetNeighborDirs(list/Dirs)
			This accepts either a single argument which is a list of directions
			or any number of arguments which are directions themselves.  This
			proc forces the grid to acknowledge neighbors in and only in the
			given directions.

*/